@@ -30,7 +30,7 @@
 #include "XrdCl/XrdClLog.hh"
 #include "XrdCl/XrdClFileSystem.hh"
 #include "XrdCl/XrdClUtils.hh"
-#include "XrdCl/XrdClRedirectorRegistry.hh"
+#include "XrdCl/XrdClDlgEnv.hh"
 #include "XrdSys/XrdSysPthread.hh"
 
 #include <stdio.h>
@@ -301,12 +301,6 @@ bool AllOptionsSupported( XrdCpConfig *config )
     return false;
   }
 
-  if( config->xRate )
-  {
-    std::cerr << "Limiting transfer rate is not yet supported" << std::endl;
-    return false;
-  }
-
   return true;
 }
 
@@ -397,46 +391,6 @@ void AdjustFileInfo( XrdCpFile *file )
   }
 };
 
-//------------------------------------------------------------------------------
-// Get a list of files and a list of directories inside a remote directory
-//------------------------------------------------------------------------------
-XrdCl::XRootDStatus GetDirList( XrdCl::FileSystem        *fs,
-                                const XrdCl::URL         &url,
-                                std::vector<std::string> *&files,
-                                std::vector<std::string> *&directories )
-{
-  using namespace XrdCl;
-  DirectoryList *list;
-  XRootDStatus   status;
-  Log *log = DefaultEnv::GetLog();
-
-  status = fs->DirList( url.GetPath(), DirListFlags::Stat, list );
-  if( !status.IsOK() )
-  {
-    log->Error( AppMsg, "Error listing directory: %s",
-                        status.ToStr().c_str());
-    return status;
-  }
-
-  for ( DirectoryList::Iterator it = list->Begin(); it != list->End(); ++it )
-  {
-    if ( (*it)->GetStatInfo()->TestFlags( StatInfo::IsDir ) )
-    {
-      std::string directory = (*it)->GetName();
-      directories->push_back( directory );
-    }
-    else
-    {
-      std::string file = (*it)->GetName();
-      files->push_back( file );
-    }
-  }
-
-  delete list;
-
-  return XRootDStatus();
-}
-
 //------------------------------------------------------------------------------
 // Recursively index all files and directories inside a remote directory
 //------------------------------------------------------------------------------
@@ -450,7 +404,8 @@ XrdCpFile *IndexRemote( XrdCl::FileSystem *fs,
   log->Debug( AppMsg, "Indexing %s", basePath.c_str() );
 
   DirectoryList *dirList = 0;
-  XRootDStatus st = fs->DirList( basePath, DirListFlags::Recursive, dirList );
+  XRootDStatus st = fs->DirList( URL( basePath ).GetPath(), DirListFlags::Recursive
+      | DirListFlags::Locate | DirListFlags::Merge, dirList );
   if( !st.IsOK() )
   {
     log->Info( AppMsg, "Failed to get directory listing for %s: %s",
@@ -465,11 +420,13 @@ XrdCpFile *IndexRemote( XrdCl::FileSystem *fs,
   for( auto itr = dirList->Begin(); itr != dirList->End(); ++itr )
   {
     DirectoryList::ListEntry *e = *itr;
+    if( e->GetStatInfo()->TestFlags( StatInfo::IsDir ) )
+      continue;
     std::string path = basePath + '/' + e->GetName();
     current = new XrdCpFile( path.c_str(), badUrl );
     if( badUrl )
     {
-      // TODO release memory
+      delete current;
       log->Error( AppMsg, "Bad URL: %s", current->Path );
       return 0;
     }
@@ -507,7 +464,7 @@ int main( int argc, char **argv )
   XrdCpConfig config( argv[0] );
   config.Config( argc, argv, XrdCpConfig::optRmtRec );
   if( !AllOptionsSupported( &config ) )
-    return 254;
+    return 50; // generic error
   ProcessCommandLineEnv( &config );
 
   //----------------------------------------------------------------------------
@@ -526,11 +483,14 @@ int main( int argc, char **argv )
   if( config.Want(XrdCpConfig::DoNoPbar) )
     progress.PrintProgressBar( false );
 
-  bool         posc      = false;
-  bool         force     = false;
-  bool         coerce    = false;
-  bool         makedir   = false;
-  bool         dynSrc    = false;
+  bool         posc          = false;
+  bool         force         = false;
+  bool         coerce        = false;
+  bool         makedir       = false;
+  bool         dynSrc        = false;
+  bool         delegate      = false;
+  bool         rmOnBadCksum  = false;
+  bool         continue_     = false;
   std::string thirdParty = "none";
 
   if( config.Want( XrdCpConfig::DoPosc ) )     posc       = true;
@@ -538,9 +498,29 @@ int main( int argc, char **argv )
   if( config.Want( XrdCpConfig::DoCoerce ) )   coerce     = true;
   if( config.Want( XrdCpConfig::DoTpc ) )      thirdParty = "first";
   if( config.Want( XrdCpConfig::DoTpcOnly ) )  thirdParty = "only";
-  if( config.Want( XrdCpConfig::DoRecurse ) )  makedir    = true;
-  if( config.Want( XrdCpConfig::DoPath    ) )  makedir    = true;
-  if( config.Want( XrdCpConfig::DoDynaSrc ) )  dynSrc     = true;
+  if( config.Want( XrdCpConfig::DoTpcDlgt ) )
+  {
+    // the env var is being set already here (we are issuing a stat
+    // inhere and we need the env var when we are establishing the
+    // connection and authenticating), but we are also setting a delegate
+    // parameter for CopyJob so it can be used on its own.
+    DlgEnv::Instance().Enable();
+    delegate = true;
+  }
+  else
+    DlgEnv::Instance().Disable();
+
+  if( config.Want( XrdCpConfig::DoRecurse ) )      makedir       = true;
+  if( config.Want( XrdCpConfig::DoPath    ) )      makedir       = true;
+  if( config.Want( XrdCpConfig::DoDynaSrc ) )      dynSrc        = true;
+  if( config.Want( XrdCpConfig::DoRmOnBadCksum ) ) rmOnBadCksum  = true;
+  if( config.Want( XrdCpConfig::DoContinue ) )     continue_     = true;
+
+  if( force && continue_ )
+  {
+    std::cerr << "Invalid argument combination: continue + force." << std::endl;
+    return 50;
+  }
 
   //----------------------------------------------------------------------------
   // Checksums
@@ -580,7 +560,7 @@ int main( int argc, char **argv )
     else
     {
       std::cerr << "Invalid parameter: " << config.CksVal << std::endl;
-      return 254;
+      return 50; // generic error
     }
   }
 
@@ -610,8 +590,11 @@ int main( int argc, char **argv )
   // Environment settings
   //----------------------------------------------------------------------------
   XrdCl::Env *env = XrdCl::DefaultEnv::GetEnv();
-  if( config.nStrm != 1 )
-    env->PutInt( "SubStreamsPerChannel", config.nStrm );
+  if( config.nStrm != 0 )
+    env->PutInt( "SubStreamsPerChannel", config.nStrm + 1 /*stands for the control stream*/ );
+
+  if( config.Want( XrdCpConfig::DoZipMtlnCksum ) )
+    env->PutInt( "ZipMtlnCksum", 1 );
 
   int chunkSize = DefaultCPChunkSize;
   env->GetInt( "CPChunkSize", chunkSize );
@@ -628,11 +611,11 @@ int main( int argc, char **argv )
     std::cerr << (int)std::numeric_limits<uint8_t>::max();
     std::cerr << " chunks in parallel. You asked for " << parallelChunks;
     std::cerr << "." << std::endl;
-    return 254;
+    return 50; // generic error
   }
 
   log->Dump( AppMsg, "Chunk size: %d, parallel chunks %d, streams: %d",
-             chunkSize, parallelChunks, config.nStrm );
+             chunkSize, parallelChunks, config.nStrm + 1 );
 
   //----------------------------------------------------------------------------
   // Build the URLs
@@ -652,8 +635,9 @@ int main( int argc, char **argv )
       char *cwd = getcwd( buf, FILENAME_MAX );
       if( !cwd )
       {
-        std::cerr <<  strerror( errno ) << std::endl;
-        return errno;
+        XRootDStatus st( stError, XProtocol::mapError( errno ), errno, strerror( errno ) );
+        std::cerr <<  st.GetErrorMessage() << std::endl;
+        return st.GetShellCode();
       }
       dest += cwd;
       dest += '/';
@@ -696,7 +680,7 @@ int main( int argc, char **argv )
   {
     std::cerr << "Multiple sources were given but target is not a directory.";
     std::cerr << std::endl;
-    return 255;
+    return 50; // generic error
   }
 
   //----------------------------------------------------------------------------
@@ -717,12 +701,12 @@ int main( int argc, char **argv )
       // Recursively index the remote directory
       //------------------------------------------------------------------------
       delete config.srcFile;
-      config.srcFile = IndexRemote( fs, source.GetURL(),
-                                    source.GetURL().size() );
+      std::string url = source.GetURL();
+      config.srcFile = IndexRemote( fs, url, url.size() );
       if ( !config.srcFile )
       {
         std::cerr << "Error indexing remote directory.";
-        return 255;
+        return 50; // generic error
       }
     }
 
@@ -755,8 +739,9 @@ int main( int argc, char **argv )
         char *cwd = getcwd( buf, FILENAME_MAX );
         if( !cwd )
         {
-          std::cerr <<  strerror( errno ) << std::endl;
-          return errno;
+          XRootDStatus st( stError, XProtocol::mapError( errno ), errno, strerror( errno ) );
+          std::cerr <<  st.GetErrorMessage() << std::endl;
+          return st.GetShellCode();
         }
         source = "file://" + std::string( cwd ) + '/' + source;
       }
@@ -768,45 +753,26 @@ int main( int argc, char **argv )
                sourceFile->Path, FileType2String( sourceFile->Protocol ),
                dest.c_str() );
 
-    //--------------------------------------------------------------------------
-    // Create a virtual redirector if it is a metalink file
-    //--------------------------------------------------------------------------
-    URL src( source );
-    if( src.IsMetalink() )
-    {
-      RedirectorRegistry &registry = RedirectorRegistry::Instance();
-      XRootDStatus st = registry.RegisterAndWait( src );
-      if( !st.IsOK() )
-      {
-        std::cerr << "RedirectorRegistry::Register " << source << " -> " << dest << ": ";
-        std::cerr << st.ToStr() << std::endl;
-        resultVect.push_back( results );
-        sourceFile = sourceFile->Next;
-        continue;
-      }
-    }
-
     //--------------------------------------------------------------------------
     // Set up the job
     //--------------------------------------------------------------------------
     std::string target = dest;
-    if( targetIsDir)
+    // if this is a recursive copy make sure we preserve the directory structure
+    if( config.Want( XrdCpConfig::DoRecurse ) )
     {
-      target = dest + "/";
-      // if it is a metalink we don't want to use the metalink name
-      if( zip )
-      {
-        target += zipFile;
-      }
-      else if( src.IsMetalink() )
-      {
-        XrdCl::RedirectorRegistry &registry = XrdCl::RedirectorRegistry::Instance();
-        VirtualRedirector *redirector = registry.Get( source );
-        target += redirector->GetTargetName();
-      }
-      else target += (sourceFile->Path+sourceFile->Doff);
+      // get the source directory
+      std::string srcDir( sourceFile->Path, sourceFile->Doff );
+      // remove the trailing slash
+      if( srcDir[srcDir.size() - 1] == '/' )
+        srcDir = srcDir.substr( 0, srcDir.size() - 1 );
+      size_t diroff = srcDir.rfind( '/' );
+      // if there is no '/' it means a directory name has been given as relative path
+      if( diroff == std::string::npos ) diroff = 0;
+      target += '/';
+      target += sourceFile->Path + diroff;
+      // remove the filename from destination path as it will be appended later anyway
+      target = target.substr( 0 , target.rfind('/') );
     }
-
     AppendCGI( target, config.dstOpq );
 
     properties.Set( "source",         source         );
@@ -825,6 +791,11 @@ int main( int argc, char **argv )
     properties.Set( "zipArchive",     zip            );
     properties.Set( "xcp",            xcp            );
     properties.Set( "xcpBlockSize",   blockSize      );
+    properties.Set( "delegate",       delegate       );
+    properties.Set( "targetIsDir",    targetIsDir    );
+    properties.Set( "xrate",          config.xRate   );
+    properties.Set( "rmOnBadCksum",   rmOnBadCksum   );
+    properties.Set( "continue",       continue_      );
 
     if( zip )
       properties.Set( "zipSource",    zipFile        );
@@ -894,6 +865,7 @@ int main( int argc, char **argv )
     return st.GetShellCode();
   }
   CleanUpResults( resultVect );
+
   return 0;
 }
 
