@@ -21,6 +21,7 @@
 #include "XrdCl/XrdClDefaultEnv.hh"
 #include "XrdCl/XrdClConstants.hh"
 #include "XrdCl/XrdClCheckSumManager.hh"
+#include "XrdCl/XrdClRedirectorRegistry.hh"
 #include "XrdNet/XrdNetAddr.hh"
 
 #include <algorithm>
@@ -32,6 +33,8 @@
 #include <locale>
 #include <map>
 #include <string>
+#include <set>
+#include <cctype>
 
 #include <sys/types.h>
 #include <dirent.h>
@@ -163,16 +166,46 @@ namespace XrdCl
       return Status( stError, errInvalidAddr );
     }
 
-    addresses.clear();
-    for( int i = 0; i < nAddrs; ++i )
-      addresses.push_back( addrs[i] );
-    delete [] addrs;
+    //--------------------------------------------------------------------------
+    // Check what are the preferences IPv6 or IPv4
+    //--------------------------------------------------------------------------
+    int preferIPv4 = DefaultPreferIPv4;
+    DefaultEnv::GetEnv()->GetInt( "PreferIPv4", preferIPv4 );
 
     //--------------------------------------------------------------------------
-    // Sort and shuffle them
+    // Partition the addresses according to the preferences
+    //
+    // The preferred IP family goes to the back as it is easier to remove
+    // items from the back of the vector
     //--------------------------------------------------------------------------
-    std::random_shuffle( addresses.begin(), addresses.end() );
-    std::sort( addresses.begin(), addresses.end(), PreferIPv6() );
+    std::vector<XrdNetAddr> result( nAddrs );
+    auto itr  = result.begin();
+    auto ritr = result.end() - 1;
+
+    for( int i = 0; i < nAddrs; ++i )
+    {
+      bool isIPv4 = addrs[i].isIPType( XrdNetAddrInfo::IPv4 ) ||
+          ( addrs[i].isIPType( XrdNetAddrInfo::IPv6 ) && addrs[i].isMapped() );
+
+      auto store = preferIPv4 ?
+                   ( isIPv4 ? ritr-- : itr++ ) :
+                   ( isIPv4 ? itr++ : ritr-- );
+
+      *store = addrs[i];
+    }
+
+    //--------------------------------------------------------------------------
+    // Shuffle each partition
+    //--------------------------------------------------------------------------
+    std::random_shuffle( result.begin(), itr );
+    std::random_shuffle( itr, result.end() );
+
+    //--------------------------------------------------------------------------
+    // Return result through output parameter
+    //--------------------------------------------------------------------------
+    addresses.swap( result );
+    delete [] addrs;
+
     return Status();
   }
 
@@ -244,17 +277,22 @@ namespace XrdCl
     delete fs;
 
     if( !st.IsOK() )
+    {
+      std::string msg = st.GetErrorMessage();
+      msg += " Got an error while querying the checksum!";
+      st.SetErrorMessage( msg );
       return st;
+    }
 
     if( !cksResponse )
-      return XRootDStatus( stError, errInternal );
+      return XRootDStatus( stError, errInternal, 0, "Got invalid response while querying the checksum!" );
 
     std::vector<std::string> elems;
     Utils::splitString( elems, cksResponse->ToString(), " " );
     delete cksResponse;
 
     if( elems.size() != 2 )
-      return XRootDStatus( stError, errInvalidResponse );
+      return XRootDStatus( stError, errInvalidResponse, 0, "Got invalid response while querying the checksum!" );
 
     if( elems[0] != checkSumType )
       return XRootDStatus( stError, errCheckSumError );
@@ -331,7 +369,7 @@ namespace XrdCl
 
     FileSystem    sourceDSFS( server );
     Buffer        queryArg; queryArg.FromString( "tpc" );
-    Buffer       *queryResponse;
+    Buffer       *queryResponse = 0;
     XRootDStatus  st;
     st = sourceDSFS.Query( QueryCode::Config, queryArg, queryResponse,
                            timeout );
@@ -343,6 +381,13 @@ namespace XrdCl
       return st;
     }
 
+    if( !queryResponse )
+    {
+      log->Error( UtilityMsg, "Cannot query source data server: empty response." );
+      st.status = stFatal;
+      return st;
+    }
+
     std::string answer = queryResponse->ToString();
     delete queryResponse;
     if( answer.length() == 1 || !isdigit( answer[0] ) || atoi(answer.c_str()) == 0)
@@ -357,6 +402,70 @@ namespace XrdCl
     return XRootDStatus();
   }
 
+  //------------------------------------------------------------------------
+  // Check if peer supports tpc / tpc lite
+  //------------------------------------------------------------------------
+  XRootDStatus Utils::CheckTPCLite( const std::string &server, uint16_t timeout )
+  {
+    Log *log = DefaultEnv::GetLog();
+    log->Debug( UtilityMsg, "Checking if the data server %s supports tpc / tpc lite",
+                server.c_str() );
+
+    FileSystem    sourceDSFS( server );
+    Buffer        queryArg; queryArg.FromString( "tpc tpcdlg" );
+    Buffer       *queryResponse = 0;
+    XRootDStatus  st;
+    st = sourceDSFS.Query( QueryCode::Config, queryArg, queryResponse,
+                           timeout );
+    if( !st.IsOK() )
+    {
+      log->Error( UtilityMsg, "Cannot query source data server %s: %s",
+                  server.c_str(), st.ToStr().c_str() );
+      st.status = stFatal;
+      return st;
+    }
+
+    if( !queryResponse )
+    {
+      log->Error( UtilityMsg, "Cannot query source data server: empty response." );
+      st.status = stFatal;
+      return st;
+    }
+
+    std::string answer = queryResponse->ToString();
+    delete queryResponse;
+
+    if( answer.empty() )
+    {
+      log->Error( UtilityMsg, "Cannot query source data server: empty response." );
+      st.status = stFatal;
+      return st;
+    }
+
+    std::vector<std::string> resp;
+    Utils::splitString( resp, answer, "\n" );
+
+    if( resp.empty() || resp[0].empty() ||
+        !isdigit( resp[0][0]) || atoi( resp[0].c_str() ) == 0 )
+    {
+      log->Debug( UtilityMsg, "Third party copy not supported at: %s",
+                  server.c_str() );
+      return XRootDStatus( stError, errNotSupported );
+    }
+
+    if( resp.size() == 1 || resp[1] == "tpcdlg" )
+    {
+      log->Debug( UtilityMsg, "TPC lite not supported at: %s",
+                  server.c_str() );
+      return XRootDStatus( stOK, suPartial );
+    }
+
+    log->Debug( UtilityMsg, "TPC lite supported at: %s",
+                server.c_str() );
+
+    return XRootDStatus();
+  }
+
   //----------------------------------------------------------------------------
   // Convert the fully qualified host name to country code
   //----------------------------------------------------------------------------
@@ -489,4 +598,146 @@ namespace XrdCl
     }
     return checksum;
   }
+
+  //----------------------------------------------------------------------------
+  // Get supported checksum types for given URL
+  //----------------------------------------------------------------------------
+  std::vector<std::string> Utils::GetSupportedCheckSums( const XrdCl::URL &url )
+  {
+    std::vector<std::string> ret;
+
+    FileSystem fs( url );
+    Buffer  arg; arg.FromString( "chksum" );
+    Buffer *resp = 0;
+    XRootDStatus st = fs.Query( QueryCode::Config, arg, resp );
+    if( st.IsOK() )
+    {
+      std::string response = resp->ToString();
+      if( response != "chksum" )
+      {
+        // we are expecting a response of format: '0:zcrc32,1:adler32'
+        std::vector<std::string> result;
+        Utils::splitString( result, response, "," );
+
+        std::vector<std::string>::iterator itr = result.begin();
+        for( ; itr != result.end(); ++itr )
+        {
+          size_t pos = itr->find( ':' );
+          if( pos == std::string::npos ) continue;
+          std::string cksname = itr->substr( pos + 1 );
+          // remove all white spaces
+          cksname.erase( std::remove_if( cksname.begin(), cksname.end(), ::isspace ), 
+                         cksname.end() );
+          ret.push_back( std::move( cksname ) );
+        }
+      }
+    }
+
+    return std::move( ret );
+  }
+
+
+  //----------------------------------------------------------------------------
+  //! Automatically infer the right checksum type
+  //----------------------------------------------------------------------------
+  std::string Utils::InferChecksumType( const XrdCl::URL &source,
+                                        const XrdCl::URL &destination,
+                                        bool              zip)
+  {
+    //--------------------------------------------------------------------------
+    // If both files are local we wont be checksumming at all
+    //--------------------------------------------------------------------------
+    if( source.IsLocalFile() && !source.IsMetalink() && destination.IsLocalFile() ) return std::string();
+
+    // checksums supported by local files
+    std::set<std::string> local_supported;
+    local_supported.insert( "adler32" );
+    local_supported.insert( "crc32" );
+    local_supported.insert( "md5" );
+    local_supported.insert( "zcrc32" );
+
+    std::vector<std::string> srccks;
+
+    if( source.IsMetalink() )
+    {
+      int useMtlnCksum = DefaultZipMtlnCksum;
+      Env *env = DefaultEnv::GetEnv();
+      env->GetInt( "ZipMtlnCksum", useMtlnCksum );
+
+      //------------------------------------------------------------------------
+      // In case of ZIP use other checksums than zcrc32 only if the user
+      // requested it explicitly.
+      //------------------------------------------------------------------------
+      if( !zip || ( zip && useMtlnCksum ) )
+      {
+        RedirectorRegistry &registry   = RedirectorRegistry::Instance();
+        VirtualRedirector  *redirector = registry.Get( source );
+        std::vector<std::string> cks = redirector->GetSupportedCheckSums();
+        srccks.insert( srccks.end(), cks.begin(), cks.end() );
+      }
+    }
+
+    if( zip )
+    {
+      //------------------------------------------------------------------------
+      // In case of ZIP we can always extract the checksum from the archive
+      //------------------------------------------------------------------------
+      srccks.push_back( "zcrc32" );
+    }
+    else if( source.GetProtocol() == "root" || source.GetProtocol() == "xroot" )
+    {
+      //------------------------------------------------------------------------
+      // If the source is a remote endpoint query the supported checksums
+      //------------------------------------------------------------------------
+      std::vector<std::string> cks = GetSupportedCheckSums( source );
+      srccks.insert( srccks.end(), cks.begin(), cks.end() );
+    }
+
+    std::vector<std::string> dstcks;
+
+    if( destination.GetProtocol() == "root" ||
+        destination.GetProtocol() == "xroot" )
+    {
+      //------------------------------------------------------------------------
+      // If the destination is a remote endpoint query the supported checksums
+      //------------------------------------------------------------------------
+      std::vector<std::string> cks = GetSupportedCheckSums( destination );
+      dstcks.insert( dstcks.end(), cks.begin(), cks.end() );
+    }
+
+    //--------------------------------------------------------------------------
+    // Now we have all the information we need, we can infer the right checksum
+    // type!!!
+    //
+    // First check if source is local
+    //--------------------------------------------------------------------------
+    if( source.IsLocalFile() && !source.IsMetalink() )
+    {
+      std::vector<std::string>::iterator itr = dstcks.begin();
+      for( ; itr != dstcks.end(); ++itr )
+        if( local_supported.count( *itr ) ) return *itr;
+      return std::string();
+    }
+
+    //--------------------------------------------------------------------------
+    // then check if destination is local
+    //--------------------------------------------------------------------------
+    if( destination.IsLocalFile() )
+    {
+      std::vector<std::string>::iterator itr = srccks.begin();
+      for( ; itr != srccks.end(); ++itr )
+        if( local_supported.count( *itr ) ) return *itr;
+      return std::string();
+    }
+
+    //--------------------------------------------------------------------------
+    // if both source and destination are remote look for a checksum that can
+    // satisfy both
+    //--------------------------------------------------------------------------
+    std::set<std::string> dst_supported( dstcks.begin(), dstcks.end() );
+    std::vector<std::string>::iterator itr = srccks.begin();
+    for( ; itr != srccks.end(); ++itr )
+      if( dst_supported.count( *itr ) ) return *itr;
+    return std::string();
+  }
 }
