@@ -28,16 +28,139 @@
 #include "XrdCl/XrdClDefaultEnv.hh"
 #include "XrdCl/XrdClLog.hh"
 #include "XrdCl/XrdClConstants.hh"
+#include "XrdCl/XrdClXRootDResponses.hh"
+#include "XrdCl/XrdClUtils.hh"
 
 #include "XrdSys/XrdSysPthread.hh"
 
+#include "XrdCks/XrdCksData.hh"
+
 #include <string>
 #include <map>
 #include <memory>
 
+#include <zlib.h>
+
 namespace XrdCl
 {
 
+struct ZipCacheError
+{
+    ZipCacheError( const XRootDStatus &status ) : status( status )
+    {
+    }
+
+    XRootDStatus status;
+};
+
+class ZipCache
+{
+  public:
+
+    ZipCache() : rawOffset( 0 ), rawSize( 0 ), totalRead( 0 )
+    {
+      strm.zalloc   = Z_NULL;
+      strm.zfree    = Z_NULL;
+      strm.opaque   = Z_NULL;
+      strm.avail_in = 0;
+      strm.next_in  = Z_NULL;
+
+      // make sure zlib doesn't look for gzip headers, in order to do so
+      // pass negative window bits !!!
+      int rc = inflateInit2( &strm, -MAX_WBITS );
+      XRootDStatus st = ToXRootDStatus( rc, "inflateInit2" );
+      if( !st.IsOK() ) throw ZipCacheError( st );
+    }
+
+    ~ZipCache()
+    {
+      inflateEnd( &strm );
+    }
+
+    XRootDStatus Input( void *inbuff, size_t insize, uint64_t rawoff )
+    {
+      // we only support streaming for compressed files
+      if( rawoff != rawOffset + rawSize )
+        return XRootDStatus( stError, errInternal );
+
+      strm.avail_in = insize;
+      strm.next_in  = (Bytef*)inbuff;
+      rawOffset     = rawoff;
+      rawSize       = insize;
+
+      return XRootDStatus();
+    }
+
+    XRootDStatus Output( void *outbuff, size_t outsize, uint64_t offset )
+    {
+      // we only support streaming for compressed files
+      if( offset != totalRead )
+        return XRootDStatus( stError, errInternal );
+
+      strm.avail_out = outsize;
+      strm.next_out  = (Bytef*)outbuff;
+
+      return XRootDStatus();
+    }
+
+    XRootDStatus Read( uint32_t &bytesRead )
+    {
+      // the available space in output buffer before inflating
+      uInt avail_before = strm.avail_out;
+
+      int rc = inflate( &strm, Z_SYNC_FLUSH );
+      XRootDStatus st = ToXRootDStatus( rc, "inflate" );
+      if( !st.IsOK() ) return st;
+
+      bytesRead = avail_before - strm.avail_out;
+      totalRead += bytesRead;
+      if( strm.avail_out ) return XRootDStatus( stOK, suPartial );
+
+      return XRootDStatus();
+    }
+
+    uint64_t NextChunkOffset()
+    {
+      return rawOffset + rawSize;
+    }
+
+  private:
+
+    XRootDStatus ToXRootDStatus( int rc, const std::string &func )
+    {
+      std::string msg = "[zlib] " + func + " : ";
+
+      switch( rc )
+      {
+        case Z_STREAM_END    :
+        case Z_OK            : return XRootDStatus();
+        case Z_BUF_ERROR     : return XRootDStatus( stOK, suContinue );
+        case Z_MEM_ERROR     : return XRootDStatus( stError, errInternal,    Z_MEM_ERROR,     msg + "not enough memory." );
+        case Z_VERSION_ERROR : return XRootDStatus( stError, errInternal,    Z_VERSION_ERROR, msg + "version mismatch." );
+        case Z_STREAM_ERROR  : return XRootDStatus( stError, errInvalidArgs, Z_STREAM_ERROR,  msg + "invalid argument." );
+        case Z_NEED_DICT     : return XRootDStatus( stError, errDataError,   Z_NEED_DICT,     msg + "need dict.");
+        case Z_DATA_ERROR    : return XRootDStatus( stError, errDataError,   Z_DATA_ERROR,    msg + "corrupted data." );
+        default              : return XRootDStatus( stError, errUnknown );
+      }
+    }
+
+    uint64_t  rawOffset; // offset of the raw data chunk in the compressed file (not archive)
+    uint32_t  rawSize;   // size of the raw data chunk
+    uint64_t  totalRead; // total number of bytes read so far
+    z_stream  strm;      // the zlib stream we will use for reading
+};
+
+
+template<typename RESP>
+struct ZipHandlerException
+{
+    ZipHandlerException( XRootDStatus *status, RESP *response ) : status( status ), response( response ) { }
+
+    XRootDStatus *status;
+    RESP         *response;
+};
+
+
 struct EOCD
 {
     EOCD( const char *buffer )
@@ -107,7 +230,7 @@ struct CDFH
       pZipVersion        = *reinterpret_cast<const uint16_t*>( buffer + 4 );
       pMinZipVersion     = *reinterpret_cast<const uint16_t*>( buffer + 6 );
       pCompressionMethod = *reinterpret_cast<const uint16_t*>( buffer + 10 );
-      pCrc32             = *reinterpret_cast<const uint32_t*>( buffer + 16 );
+      pZCRC32             = *reinterpret_cast<const uint32_t*>( buffer + 16 );
       pCompressedSize    = *reinterpret_cast<const uint32_t*>( buffer + 20 );
       pUncompressedSize  = *reinterpret_cast<const uint32_t*>( buffer + 24 );
       pDiskNb            = *reinterpret_cast<const uint16_t*>( buffer + 34 );
@@ -119,17 +242,99 @@ struct CDFH
 
       pFilename = std::string( buffer + 46, filenameLength );
 
+      // now parse the 'extra' (may contain the zip64 extension to CDFH)
+      ParseExtra( buffer + 46 + filenameLength, extraLength );
+
       pCdfhSize = kCdfhBaseSize + filenameLength +extraLength + commentLength;
     }
 
+    void ParseExtra( const char *buffer, uint16_t length)
+    {
+      static const uint32_t ovrflw32 = 0xffffffff;
+      static const uint16_t ovrflw16 = 0xffff;
+      static const uint16_t ZIP64_EXTENSION_SIGN = 0x0001;
+
+      uint16_t exsize = 0;
+      uint64_t *ovrflw[3] = {0, 0, 0};
+      uint8_t count = 0;
+
+      // check if compressed size is overflown
+      if( pCompressedSize == ovrflw32)
+      {
+        ovrflw[count] = &pCompressedSize;
+        ++count;
+        exsize += sizeof( uint64_t );
+      }
+
+      // check if original size is overflown
+      if( pUncompressedSize == ovrflw32 )
+      {
+        ovrflw[count] = &pUncompressedSize;
+        ++count;
+        exsize += sizeof( uint64_t );
+      }
+
+      // check if offset is overflown
+      if( pOffset == ovrflw32 )
+      {
+        ovrflw[count] = &pOffset;
+        ++count;
+        exsize += sizeof( uint64_t );
+      }
+
+      // check if number of disks is overflown
+      if( pDiskNb == ovrflw16 )
+        exsize += sizeof( uint32_t );
+
+      // if the expected size of ZIP64 extension is 0 we
+      // can skip parsing of 'extra'
+      if( exsize == 0 ) return;
+
+      // Parse the extra part
+      const char *end = buffer + length;
+      uint16_t signature = 0;
+      uint16_t blksize   = 0;
+      while( buffer < end )
+      {
+        signature = *reinterpret_cast<const uint16_t*>( buffer );
+        buffer += sizeof( uint16_t );
+        blksize   = *reinterpret_cast<const uint16_t*>( buffer );
+        buffer += sizeof( uint16_t );
+
+        // is it ZIP64 extension
+        if( signature == ZIP64_EXTENSION_SIGN )
+        {
+          if( blksize != exsize )
+            throw ZipHandlerException<AnyObject>( new XRootDStatus( stError, errDataError, 0, "Wrong size of ZIP64 extension!" ), 0 );
+
+          for( uint8_t j = 0; j < count; ++j )
+          {
+            uint64_t value = *reinterpret_cast<const uint64_t*>( buffer );
+            buffer += sizeof( uint64_t );
+            exsize -= sizeof( uint64_t );
+            *ovrflw[j] = value;
+          }
+
+          if( exsize == sizeof( uint32_t ) )
+          {
+            pDiskNb = *reinterpret_cast<const uint32_t*>( buffer );
+            buffer += sizeof( uint32_t );
+          }
+
+          break;
+        }
+        buffer += blksize;
+      }
+    }
+
     uint16_t    pZipVersion;
     uint16_t    pMinZipVersion;
     uint16_t    pCompressionMethod;
-    uint32_t    pCrc32;
-    uint32_t    pCompressedSize;
-    uint32_t    pUncompressedSize;
-    uint16_t    pDiskNb;
-    uint32_t    pOffset;
+    uint32_t    pZCRC32;
+    uint64_t    pCompressedSize;
+    uint64_t    pUncompressedSize;
+    uint32_t    pDiskNb;
+    uint64_t    pOffset;
     std::string pFilename;
     uint16_t    pCdfhSize;
 
@@ -142,7 +347,7 @@ class ZipArchiveReaderImpl
 {
   public:
 
-    ZipArchiveReaderImpl() : pArchiveSize( 0 ), pRefCount( 1 ), pOpen( false ) { }
+    ZipArchiveReaderImpl( File &archive ) : pArchive( archive ), pArchiveSize( 0 ), pRefCount( 1 ), pOpen( false ) { }
 
     ZipArchiveReaderImpl* Self()
     {
@@ -174,6 +379,20 @@ class ZipArchiveReaderImpl
 
     XRootDStatus Read( const std::string &filename, uint64_t relativeOffset, uint32_t size, void *buffer, ResponseHandler *userHandler, uint16_t timeout = 0 );
 
+    XRootDStatus ZCRC32( const std::string &filename, std::string &checksum );
+
+    XRootDStatus ZCRC32( std::string &checksum );
+
+    XRootDStatus Read( uint64_t relativeOffset, uint32_t size, void *buffer, ResponseHandler *userHandler, uint16_t timeout = 0 )
+    {
+      if( pBoundFile.empty() )
+        return XRootDStatus( stError, errInvalidOp );
+
+      return Read( pBoundFile, relativeOffset, size, buffer, userHandler, timeout );
+    }
+
+    DirectoryList* List();
+
     XRootDStatus Close( ResponseHandler *handler, uint16_t timeout )
     {
       XRootDStatus st = pArchive.Close( handler, timeout );
@@ -185,17 +404,20 @@ class ZipArchiveReaderImpl
       return st;
     }
 
-    bool SetProperty( const std::string &name, const std::string &value )
-    {
-      return pArchive.SetProperty( name, value );
-    }
-
-    XRootDStatus GetSize( const std::string & filename, uint32_t &size ) const
+    XRootDStatus GetSize( const std::string & filename, uint64_t &size ) const
     {
       std::map<std::string, size_t>::const_iterator it = pFileToCdfh.find( filename );
       if( it == pFileToCdfh.end() ) return XRootDStatus( stError, errNotFound );
       CDFH *cdfh = pCdRecords[it->second];
-      size = cdfh->pCompressionMethod ? cdfh->pCompressedSize : cdfh->pUncompressedSize;
+      size = cdfh->pCompressionMethod ? cdfh->pUncompressedSize : cdfh->pCompressedSize;
+      return XRootDStatus();
+    }
+
+    XRootDStatus Bind( const std::string &filename )
+    {
+      std::map<std::string, size_t>::const_iterator it = pFileToCdfh.find( filename );
+      if( it == pFileToCdfh.end() ) return XRootDStatus( stError, errNotFound );
+      pBoundFile = filename;
       return XRootDStatus();
     }
 
@@ -278,6 +500,8 @@ class ZipArchiveReaderImpl
         delete *it;
       pCdRecords.clear();
       pFileToCdfh.clear();
+
+      pBoundFile.erase();
     }
 
     ~ZipArchiveReaderImpl()
@@ -294,26 +518,18 @@ class ZipArchiveReaderImpl
       }
     }
 
-    File                           pArchive;
-    std::string                    pFilename;
-    uint64_t                       pArchiveSize;
-    std::unique_ptr<char[]>        pBuffer;
-    std::unique_ptr<EOCD>          pEocd;
-    std::unique_ptr<ZIP64_EOCD>    pZip64Eocd;
-    std::vector<CDFH*>             pCdRecords;
-    std::map<std::string, size_t>  pFileToCdfh;
-    mutable XrdSysMutex            pMutex;
-    size_t                         pRefCount;
-    bool                           pOpen;
-};
-
-template<typename RESP>
-struct ZipHandlerException
-{
-    ZipHandlerException( XRootDStatus *status, RESP *response ) : status( status ), response( response ) { }
-
-    XRootDStatus *status;
-    RESP         *response;
+    File                             &pArchive;
+    uint64_t                          pArchiveSize;
+    std::unique_ptr<char[]>           pBuffer;
+    std::unique_ptr<EOCD>             pEocd;
+    std::unique_ptr<ZIP64_EOCD>       pZip64Eocd;
+    std::vector<CDFH*>                pCdRecords;
+    std::map<std::string, size_t>     pFileToCdfh;
+    mutable XrdSysMutex               pMutex;
+    size_t                            pRefCount;
+    bool                              pOpen;
+    std::string                       pBoundFile;
+    std::map<std::string, ZipCache>   pCompressed;
 };
 
 
@@ -566,7 +782,70 @@ class ZipReadHandler : public ZipHandlerBase<ChunkInfo>
 };
 
 
-ZipArchiveReader::ZipArchiveReader() : pImpl( new ZipArchiveReaderImpl() )
+class ZipReadCompressedHandler : public ZipHandlerBase<ChunkInfo>
+{
+  public:
+
+    ZipReadCompressedHandler( ZipCache             &cache,
+                              uint64_t              relativeOffet,
+                              uint64_t              rawOffset,
+                              void                 *userBuffer,
+                              uint32_t              userSize,
+                              ZipArchiveReaderImpl *impl,
+                              ResponseHandler      *userHandler ) : ZipHandlerBase<ChunkInfo>( impl, userHandler ),
+                                                                    pCache( cache ),
+                                                                    pUserOffset( relativeOffet ),
+                                                                    pRawOffset( rawOffset ),
+                                                                    pUserBuffer( userBuffer ),
+                                                                    pUserSize( userSize )
+    {
+
+    }
+
+    virtual void HandleResponseImpl( XRootDStatus *status, ChunkInfo *response )
+    {
+      XRootDStatus st = pCache.Input( response->buffer, response->length, pRawOffset );
+      if( !st.IsOK() )
+      {
+        DeleteArgs( status, response );
+        if( pUserHandler ) pUserHandler->HandleResponse( new XRootDStatus( st ), 0 );
+        return;
+      }
+
+      // at this point we can be sure that all the needed data are in the cache
+      // (we requested as much data as the user asked for so in the worst case
+      // we have exactly as much data as the user needs, most likely we have
+      // more because the data are compressed)
+      uint32_t bytesRead = 0;
+      st = pCache.Read( bytesRead );
+      if( !st.IsOK() )
+      {
+        DeleteArgs( status, response );
+        if( pUserHandler ) pUserHandler->HandleResponse( new XRootDStatus( st ), 0 );
+        return;
+      }
+
+      // prepare the response for the end-user
+      response->buffer = pUserBuffer;
+      response->length = pUserSize;
+      response->offset = pUserOffset;
+
+      if( pUserHandler ) pUserHandler->HandleResponse( status, PkgResp( response ) );
+      else
+        DeleteArgs( status, response );
+    }
+
+  private:
+
+    ZipCache &pCache;
+    uint64_t  pUserOffset;
+    uint64_t  pRawOffset;
+    void     *pUserBuffer;
+    uint32_t  pUserSize;
+};
+
+
+ZipArchiveReader::ZipArchiveReader( File &archive ) : pImpl( new ZipArchiveReaderImpl( archive ) )
 {
 
 }
@@ -611,8 +890,22 @@ XRootDStatus ZipArchiveReaderImpl::StatArchive( ResponseHandler *userHandler )
   // just to be on the safe side
   ClearRecords();
 
+  // create a stat handler
   StatArchiveHandler *handler = new StatArchiveHandler( this, userHandler );
-  XRootDStatus st = pArchive.Stat( true, handler );
+
+  // let's check if we got a stat together with the open response
+  StatInfo *response = 0;
+  XRootDStatus st = pArchive.Stat( false, response ); // always returns stOK
+  if( st.IsOK() && response )
+  {
+    AnyObject *resp = new AnyObject();
+    resp->Set( response );
+    handler->HandleResponse( new XRootDStatus(), resp ); //this will deallocate the handler memory
+    return XRootDStatus();
+  }
+
+  // if we don't have the stat yet, we need to issue a stat request
+  st = pArchive.Stat( true, handler );
   if( !st.IsOK() ) delete handler;
   return st;
 }
@@ -710,6 +1003,54 @@ XRootDStatus ZipArchiveReader::Read( const std::string &filename, uint64_t offse
   return status;
 }
 
+//------------------------------------------------------------------------
+// Bounds the reader to a file inside the archive.
+//------------------------------------------------------------------------
+XRootDStatus ZipArchiveReader::Bind( const std::string &filename )
+{
+  return pImpl->Bind( filename );
+}
+
+//------------------------------------------------------------------------
+// Async bound read.
+//------------------------------------------------------------------------
+XRootDStatus ZipArchiveReader::Read( uint64_t offset, uint32_t size, void *buffer, ResponseHandler *handler, uint16_t timeout )
+{
+  return pImpl->Read( offset, size, buffer, handler, timeout );
+}
+
+//------------------------------------------------------------------------
+// Sync bound read.
+//------------------------------------------------------------------------
+XRootDStatus ZipArchiveReader::Read( uint64_t offset, uint32_t size, void *buffer, uint32_t &bytesRead, uint16_t timeout )
+{
+  SyncResponseHandler handler;
+  Status st = Read( offset, size, buffer, &handler, timeout );
+  if( !st.IsOK() )
+    return st;
+
+  ChunkInfo *chunkInfo = 0;
+  XRootDStatus status = MessageUtils::WaitForResponse( &handler, chunkInfo );
+  if( status.IsOK() )
+  {
+    bytesRead = chunkInfo->length;
+    delete chunkInfo;
+  }
+  return status;
+}
+
+//------------------------------------------------------------------------
+// Sync list
+//------------------------------------------------------------------------
+XRootDStatus ZipArchiveReader::List( DirectoryList *&list )
+{
+  if( !pImpl->IsOpen() )
+    return XRootDStatus( stError, errInvalidOp );
+
+  list = pImpl->List();
+  return XRootDStatus();
+}
+
 XRootDStatus ZipArchiveReaderImpl::Read( const std::string &filename, uint64_t relativeOffset, uint32_t size, void *buffer, ResponseHandler *userHandler, uint16_t timeout )
 {
   if( !pArchive.IsOpen() ) return XRootDStatus( stError, errInvalidOp, errInvalidOp, "Archive not opened." );
@@ -718,6 +1059,10 @@ XRootDStatus ZipArchiveReaderImpl::Read( const std::string &filename, uint64_t r
   if( cditr == pFileToCdfh.end() ) return XRootDStatus( stError, errNotFound, errNotFound, "File not found." );
   CDFH *cdfh = pCdRecords[cditr->second];
 
+  // check if the file is compressed, for now we only support uncompressed files!
+  if( cdfh->pCompressionMethod != 0 && cdfh->pCompressionMethod != Z_DEFLATED )
+    return XRootDStatus( stError, errNotSupported, 0, "The compression algorithm is not supported!" );
+
   // Now the problem is that at the beginning of our
   // file there is the Local-file-header, which size
   // is not known because of the variable size 'extra'
@@ -725,12 +1070,71 @@ XRootDStatus ZipArchiveReaderImpl::Read( const std::string &filename, uint64_t r
   // record and shift it by the file size.
   // The next record is either the next LFH (next file)
   // or the start of the Central-directory.
-  uint64_t nextRecordOffset = ( cditr->second + 1 < pCdRecords.size() ) ? pCdRecords[cditr->second + 1]->pOffset : pEocd->pCdOffset;
-  uint32_t fileSize = cdfh->pCompressionMethod ? cdfh->pCompressedSize : cdfh->pUncompressedSize;
-  uint64_t offset = nextRecordOffset - fileSize + relativeOffset;
-  uint32_t sizeTillEnd = fileSize - relativeOffset;
+  uint64_t cdOffset = pZip64Eocd ? pZip64Eocd->pCdOffset : pEocd->pCdOffset;
+  uint64_t nextRecordOffset = ( cditr->second + 1 < pCdRecords.size() ) ? pCdRecords[cditr->second + 1]->pOffset : cdOffset;
+  uint64_t filesize  = cdfh->pCompressedSize;
+  uint64_t fileoff  = nextRecordOffset - filesize;
+  uint64_t offset   = fileoff + relativeOffset;
+  uint64_t sizeTillEnd = cdfh->pUncompressedSize - relativeOffset;
   if( size > sizeTillEnd ) size = sizeTillEnd;
 
+  // if it is a compressed file use ZIP cache to read from the file
+  if( cdfh->pCompressionMethod == Z_DEFLATED )
+  {
+    // check if respective ZIP cache exists
+    bool recent = pCompressed.find( filename ) == pCompressed.end();
+    // if the entry does not exist, it will be created using
+    // default constructor
+    ZipCache &cache = pCompressed[filename];
+    // if we have the whole ZIP archive we can populate the cache
+    // straight away
+    if( recent && pBuffer)
+    {
+      XRootDStatus st = cache.Input( pBuffer.get() + offset, filesize - fileoff, relativeOffset );
+      if( !st.IsOK() ) return st;
+    }
+
+    XRootDStatus st = cache.Output( buffer, size, relativeOffset );
+
+    uint32_t bytesRead = 0;
+    st = cache.Read( bytesRead );
+
+    // propagate errors to the end-user
+    if( !st.IsOK() ) return st;
+
+    if( st.code == suPartial )
+    {
+      // the raw offset of the next chunk within the file
+      uint64_t rawOffset = cache.NextChunkOffset();
+      // if this is the first time we are setting an input chunk
+      // use the user-specified offset
+      if( !rawOffset )
+        rawOffset = relativeOffset;
+      // size of the next chunk of raw (compressed) data
+      uint32_t chunkSize = size;
+      // make sure we are not reading passed the end of the file
+      if( rawOffset + chunkSize > filesize )
+        chunkSize = filesize - rawOffset;
+      // allocate the buffer for the compressed data
+      pBuffer.reset( new char[chunkSize] );
+      ZipReadCompressedHandler *handler = new ZipReadCompressedHandler( cache, relativeOffset, rawOffset, buffer, size, this, userHandler );
+      st = pArchive.Read( fileoff + rawOffset, chunkSize, pBuffer.get(), handler, timeout );
+      if( !st.IsOK() ) delete handler;
+      return st;
+    }
+
+    if( userHandler )
+    {
+      XRootDStatus *st   = new XRootDStatus();
+      AnyObject    *resp = new AnyObject();
+      ChunkInfo    *info = new ChunkInfo( relativeOffset, size, buffer );
+      resp->Set( info );
+      userHandler->HandleResponse( st, resp );
+    }
+
+    return XRootDStatus();
+  }
+
   // check if we have the whole file in our local buffer
   if( pBuffer )
   {
@@ -760,6 +1164,59 @@ XRootDStatus ZipArchiveReaderImpl::Read( const std::string &filename, uint64_t r
   return st;
 }
 
+XRootDStatus ZipArchiveReaderImpl::ZCRC32( const std::string &filename, std::string &checksum )
+{
+  if( !pArchive.IsOpen() ) return XRootDStatus( stError, errInvalidOp, errInvalidOp, "Archive not opened." );
+
+  std::map<std::string, size_t>::iterator cditr = pFileToCdfh.find( filename );
+  if( cditr == pFileToCdfh.end() ) return XRootDStatus( stError, errNotFound, errNotFound, "File not found." );
+  CDFH *cdfh = pCdRecords[cditr->second];
+  XrdCksData ckSum;
+  ckSum.Set( "zcrc32" );
+  ckSum.Set( reinterpret_cast<void*>( &cdfh->pZCRC32 ), sizeof( uint32_t ) );
+
+  char *cksBuffer = new char[265];
+  ckSum.Get( cksBuffer, 256 );
+  checksum  = "zcrc32:";
+  checksum += Utils::NormalizeChecksum( "zcrc32", cksBuffer );
+  delete [] cksBuffer;
+  return XRootDStatus();
+}
+
+XRootDStatus ZipArchiveReaderImpl::ZCRC32( std::string &checksum )
+{
+  return ZCRC32( pBoundFile, checksum );
+}
+
+DirectoryList* ZipArchiveReaderImpl::List()
+{
+  std::string value;
+  pArchive.GetProperty( "LastURL", value );
+  URL url( value );
+
+  StatInfo *infoptr = 0;
+  XRootDStatus st = pArchive.Stat( false, infoptr );
+  std::unique_ptr<StatInfo> info( infoptr );
+
+  DirectoryList *list = new DirectoryList();
+  list->SetParentName( url.GetPath() );
+
+  auto itr = pCdRecords.begin();
+  for( ; itr != pCdRecords.end() ; ++itr )
+  {
+    CDFH *cdfh = *itr;
+    StatInfo *entry_info = new StatInfo( info->GetId(),
+                                         cdfh->pCdfhSize,
+                                         info->GetFlags() & ( ~StatInfo::IsWritable ), // make sure it is not listed as writable
+                                         info->GetModTime() );
+    DirectoryList::ListEntry *entry =
+        new DirectoryList::ListEntry( url.GetHostId(), cdfh->pFilename, entry_info );
+    list->Add( entry );
+  }
+
+  return list;
+}
+
 XRootDStatus ZipArchiveReader::Close( ResponseHandler *handler, uint16_t timeout )
 {
   return pImpl->Close( handler, timeout );
@@ -775,12 +1232,7 @@ XRootDStatus ZipArchiveReader::Close( uint16_t timeout )
   return MessageUtils::WaitForStatus( &handler );
 }
 
-bool ZipArchiveReader::SetProperty( const std::string &name, const std::string &value )
-{
-  return pImpl->SetProperty( name, value );
-}
-
-XRootDStatus ZipArchiveReader::GetSize( const std::string &filename, uint32_t &size ) const
+XRootDStatus ZipArchiveReader::GetSize( const std::string &filename, uint64_t &size ) const
 {
   return pImpl->GetSize( filename, size );
 }
@@ -790,4 +1242,14 @@ bool ZipArchiveReader::IsOpen() const
   return pImpl->IsOpen();
 }
 
+XRootDStatus ZipArchiveReader::ZCRC32( const std::string &filename, std::string &checksum )
+{
+  return pImpl->ZCRC32( filename, checksum );
+}
+
+XRootDStatus ZipArchiveReader::ZCRC32( std::string &checksum )
+{
+  return pImpl->ZCRC32( checksum );
+}
+
 } /* namespace XrdCl */
