@@ -50,7 +50,7 @@
 #ifndef s6_addr32
 #if   defined(__solaris__)
 #define s6_addr32 _S6_un._S6_u32
-#elif defined(__APPLE__)
+#elif defined(__APPLE__) || defined(__FreeBSD__)
 #define s6_addr32 __u6_addr.__u6_addr32
 #endif
 #endif
@@ -85,6 +85,7 @@ struct addrinfo   *XrdNetAddr::huntHintsUDP = XrdNetAddr::Hints(2, SOCK_DGRAM);
 // The following must be initialzed after all of the hint structures!
 //
 bool               XrdNetAddr::useIPV4      = OnlyIPV4();
+bool               XrdNetAddr::dynDNS       = false;
 
 /******************************************************************************/
 /*                           C o n s t r u c t o r                            */
@@ -267,6 +268,8 @@ const char *XrdNetAddr::Set(const char *hSpec, int pNum)
             n = getaddrinfo(iP, 0, hostHints, &rP);
             if (n || !rP)
                {if (rP) freeaddrinfo(rP);
+                if (n == EAI_NONAME && dynDNS)
+                   return "Dynamic name or service not yet registered";
                 return (n ? gai_strerror(n) : "host not found");
                }
             memcpy(&IP.Addr, rP->ai_addr, rP->ai_addrlen);
@@ -357,11 +360,15 @@ const char *XrdNetAddr::Set(const char *hSpec, int &numIP, int maxIP,
 
 const char *XrdNetAddr::Set(const struct sockaddr *sockP, int sockFD)
 {
+// Make sure we won't loose any bits of sockFD (we should use an int)
+//
+   if (sockFD >=0 && (sockFD & 0xffff0000) != 0) return "FD is out of range";
+
 // Clear translation if set
 //
    if (hostName)             {free(hostName);  hostName = 0;}
    if (sockAddr != &IP.Addr) {delete unixPipe; sockAddr = &IP.Addr;}
-   sockNum = sockFD;
+   sockNum = static_cast<unsigned short>(sockFD);
 
 // Copy the address based on address family
 //
@@ -395,12 +402,16 @@ const char *XrdNetAddr::Set(int sockFD, bool peer)
 {
    int rc;
 
+// Make sure we won't loose any bits of sockFD (we should use an int)
+//
+   if ((sockFD & 0xffff0000) != 0) return "FD is out of range";
+
 // Clear translation if set
 //
    if (hostName)             {free(hostName);  hostName = 0;}
    if (sockAddr != &IP.Addr) {delete unixPipe; sockAddr = &IP.Addr;}
    addrSize = sizeof(sockaddr_in6);
-   sockNum = sockFD;
+   sockNum = static_cast<unsigned short>(sockFD);
 
 // Get the address on the appropriate side of this socket
 //
@@ -451,7 +462,7 @@ const char *XrdNetAddr::Set(struct addrinfo *rP, int Port, bool mapit)
    hostName = (rP->ai_canonname ? strdup(rP->ai_canonname) : 0);
    if (sockAddr != &IP.Addr) {delete unixPipe; sockAddr = &IP.Addr;}
    IP.v6.sin6_port = htons(static_cast<short>(Port));
-   sockNum = -1;
+   sockNum = 0;
    return 0;
 }
 
@@ -469,6 +480,12 @@ void XrdNetAddr::SetCache(int keeptime)
    dnsCache = (keeptime > 0 ? &theCache : 0);
 }
 
+/******************************************************************************/
+/*                             S e t D y n D N S                              */
+/******************************************************************************/
+  
+void XrdNetAddr::SetDynDNS(bool onoff) {dynDNS = onoff;}
+
 /******************************************************************************/
 /*                               S e t I P V 4                                */
 /******************************************************************************/
