@@ -46,6 +46,7 @@
 
 #include "XrdOfs/XrdOfs.hh"
 #include "XrdOfs/XrdOfsConfigPI.hh"
+#include "XrdOfs/XrdOfsPrepare.hh"
 
 #include "XrdOss/XrdOss.hh"
 
@@ -65,7 +66,8 @@
   
 namespace
 {
-const char *drctv[] = {"xattrlib", "authlib", "ckslib", "cmslib", "osslib"};
+const char *drctv[] = {"xattrlib", "authlib", "ckslib", "cmslib",
+                         "osslib", "preplib"};
 
 const char *nullParms = 0;
 }
@@ -86,10 +88,10 @@ XrdVERSIONINFOREF(XrdOfs);
   
 XrdOfsConfigPI::XrdOfsConfigPI(const char  *cfn,  XrdOucStream   *cfgP,
                                XrdSysError *errP, XrdVersionInfo *verP)
-                 : autPI(0), cksPI(0), cmsPI(0), ossPI(0), urVer(verP),
+                 : autPI(0), cksPI(0), cmsPI(0), prpPI(0), ossPI(0), urVer(verP),
                    Config(cfgP),  Eroute(errP), CksConfig(0), ConfigFN(cfn),
                    CksAlg(0), CksRdsz(0), ossXAttr(false), ossCksio(false),
-                   Loaded(false), LoadOK(false)
+                   prpAuth(true), Loaded(false), LoadOK(false), cksLcl(false)
 {
    int rc;
 
@@ -179,7 +181,7 @@ void   XrdOfsConfigPI::Display()
 /*                                  L o a d                                   */
 /******************************************************************************/
   
-bool XrdOfsConfigPI::Load(int loadLib, XrdOucEnv *envP)
+bool XrdOfsConfigPI::Load(int loadLib, XrdOfs *ofsP, XrdOucEnv *envP)
 {
    extern XrdOss *XrdOssGetSS(XrdSysLogger *, const char *, const char *,
                               const char   *, XrdOucEnv  *, XrdVersionInfo &);
@@ -232,6 +234,10 @@ bool XrdOfsConfigPI::Load(int loadLib, XrdOucEnv *envP)
 //
    if (DO_LOAD(theCmsLib) && !SetupCms()) return false;
 
+// Setup the prepare plugin if need be
+//
+   if (DO_LOAD(thePrpLib) && !SetupPrp(ofsP, envP)) return false;
+
 // All done
 //
    LoadOK = true;
@@ -276,8 +282,11 @@ bool XrdOfsConfigPI::Parse(TheLib what)
                           break;
           case theAutLib: break;
           case theCksLib: if (CksConfig)
-                             {if (CksConfig->ParseLib(*Config)) return false;
-                              RepLib(theCksLib, CksConfig->ManLib(), nullParms);
+                             {int libType;
+                              if (CksConfig->ParseLib(*Config, libType))
+                                 return false;
+                              if (libType) cksLcl = libType == 1;
+                              RepLib(theCksLib, CksConfig->ManLib(), nullParms, false);
                               return true;
                              }
                           Eroute->Emsg("Config", "Checksum version error!");
@@ -287,6 +296,8 @@ bool XrdOfsConfigPI::Parse(TheLib what)
                           break;
           case theOssLib: return ParseOssLib();
                           break;
+          case thePrpLib: return ParsePrpLib();
+                          break;
           default:        Eroute->Emsg("Config", "Invalid plugin Parse() call");
                           return false;
                           break;
@@ -380,6 +391,52 @@ bool XrdOfsConfigPI::ParseOssLib()
    return RepLib(theOssLib, val);
 }
 
+/******************************************************************************/
+/* Private:                  P a r s e P r p L i b                            */
+/******************************************************************************/
+  
+/* Function: ParsePrpLib
+
+   Purpose:  To parse the directive: preplib [<opts>] <path> [<parms>]
+             <opts>: [+noauth]
+
+             +noauth   do not apply authorization to path list.
+             <path>    the path of the prepare library to be used.
+             <parms>   optional parms to be passed
+
+  Output: true upon success or false upon failure.
+*/
+
+bool XrdOfsConfigPI::ParsePrpLib()
+{
+   char *val, oBuff[80];
+   int   oI = PIX(thePrpLib);
+
+// Reset to defaults
+//
+    prpAuth = true;
+    if (LP[oI].opts) {free(LP[oI].opts); LP[oI].opts = 0;}
+    *oBuff = 0;
+
+// Get the path and parms, and process keywords
+//
+   while((val = Config->GetWord()))
+        {     if (!strcmp("+noauth",  val))
+                 {if (prpAuth) strcat(oBuff, "+noauth "); prpAuth = false;}
+         else break;
+        }
+
+// Check if we a library path
+//
+   if (!val || !val[0])
+      {Eroute->Emsg("Config", "preplib not specified"); return false;}
+
+// Record the path and parameters
+//
+   if (*oBuff) LP[oI].opts = strdup(oBuff);
+   return RepLib(thePrpLib, val);
+}
+
 /******************************************************************************/
 /*                                P l u g i n                                 */
 /******************************************************************************/
@@ -393,15 +450,24 @@ bool   XrdOfsConfigPI::Plugin(XrdCks          *&piP)
 bool   XrdOfsConfigPI::Plugin(XrdCmsClient_t   &piP)
 {      return (piP = cmsPI) != 0;}
 
+bool   XrdOfsConfigPI::Plugin(XrdOfsPrepare   *&piP)
+{      return (piP = prpPI) != 0;}
+
 bool   XrdOfsConfigPI::Plugin(XrdOss          *&piP)
 {      return (piP = ossPI) != 0;}
 
+/******************************************************************************/
+/*                              P r e p A u t h                               */
+/******************************************************************************/
+
+bool   XrdOfsConfigPI::PrepAuth() {return prpAuth;}
+  
 /******************************************************************************/
 /* Private:                       R e p L i b                                 */
 /******************************************************************************/
   
 bool XrdOfsConfigPI::RepLib(XrdOfsConfigPI::TheLib what,
-                            const char *newLib, const char *newParms)
+                            const char *newLib, const char *newParms, bool parseParms)
 {
    const char *parmP;
    char  parms[2048];
@@ -424,7 +490,7 @@ bool XrdOfsConfigPI::RepLib(XrdOfsConfigPI::TheLib what,
 //
    if (newParms) parmP = (newParms == nullParms ? 0 : newParms);
       else {*parms = 0; parmP = parms;
-            if (!Config->GetRest(parms, sizeof(parms)))
+            if (parseParms && !Config->GetRest(parms, sizeof(parms)))
                {Eroute->Emsg("Config", drctv[xLib], "parameters too long");
                 return false;
                }
@@ -530,3 +596,32 @@ bool XrdOfsConfigPI::SetupCms()
       }
    return true;
 }
+
+/******************************************************************************/
+/*                              S e t u p P r p                               */
+/******************************************************************************/
+  
+bool XrdOfsConfigPI::SetupPrp(XrdOfs *ofsP, XrdOucEnv *envP)
+{
+   XrdOfsgetPrepare_t ep = 0;
+   char *PrpLib   = LP[PIX(thePrpLib)].lib;
+   char *PrpParms = LP[PIX(thePrpLib)].parms;
+
+// Load the plugin if we have to
+//
+   if (LP[PIX(thePrpLib)].lib)
+      {XrdOucPinLoader myLib(Eroute, urVer, "preplib", PrpLib);
+       ep = (XrdOfsgetPrepare_t)(myLib.Resolve("XrdOfsgetPrepare"));
+       if (!ep) return false;
+       if (strcmp(PrpLib, myLib.Path()))
+          {free(PrpLib);
+           PrpLib = LP[PIX(thePrpLib)].lib = strdup(myLib.Path());
+          }
+      }
+
+// Get the Object now
+//
+   if (ep)
+      return 0 != (prpPI = ep(Eroute, ConfigFN, PrpParms, ofsP, ossPI, envP));
+   return true;
+}
