@@ -29,6 +29,7 @@
 /******************************************************************************/
 
 #include <errno.h>
+#include <iostream>
 #include <stdio.h>
 
 #include "XrdCl/XrdClDefaultEnv.hh"
@@ -39,15 +40,18 @@
 #include "XrdOuc/XrdOucPsx.hh"
 #include "XrdOuc/XrdOucTList.hh"
 
+#include "XrdPosix/XrdPosixCache.hh"
 #include "XrdPosix/XrdPosixCacheBC.hh"
 #include "XrdPosix/XrdPosixConfig.hh"
 #include "XrdPosix/XrdPosixFileRH.hh"
+#include "XrdPosix/XrdPosixInfo.hh"
 #include "XrdPosix/XrdPosixMap.hh"
 #include "XrdPosix/XrdPosixPrepIO.hh"
 #include "XrdPosix/XrdPosixTrace.hh"
 #include "XrdPosix/XrdPosixXrootd.hh"
 #include "XrdPosix/XrdPosixXrootdPath.hh"
 
+#include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysTrace.hh"
 
 /******************************************************************************/
@@ -63,6 +67,7 @@ extern XrdOucCache2              *myCache2;
 extern XrdOucName2Name           *theN2N;
 extern XrdCl::DirListFlags::Flags dlFlag;
 extern XrdSysLogger              *theLogger;
+extern XrdSysError               *eDest;
 extern XrdSysTrace                Trace;
 extern int                        ddInterval;
 extern int                        ddMaxTries;
@@ -86,7 +91,33 @@ void XrdPosixConfig::EnvInfo(XrdOucEnv &theEnv)
 }
 
 /******************************************************************************/
-/*                               i n i t E n v                                */
+/* Private:                      i n i t C C M                                */
+/******************************************************************************/
+
+bool XrdPosixConfig::initCCM(XrdOucPsx &parms)
+{
+   static XrdPosixCache pCache;
+   const char *eTxt = "Unable to initialize cache context manager in";
+   const char *mPath;
+
+// Initialize the cache context manager
+//
+   if ((*parms.initCCM)(pCache, parms.theLogger, parms.configFN,
+                        parms.CCMInfo(mPath), parms.theEnv)) return true;
+
+// Issue error message and return failure
+//
+   if (parms.theLogger)
+      {XrdSysError eDest(parms.theLogger, "Posix");
+       eDest.Emsg("InitCCM", eTxt, mPath);
+      } else {
+       std::cerr <<"Posix_InitCCM: " <<eTxt <<' ' <<mPath <<std::endl;
+      }
+    return false;
+}
+
+/******************************************************************************/
+/* Private:                      i n i t E n v                                */
 /******************************************************************************/
 
 // Parse options specified as a cgi string (i.e. var=val&var=val&...). Vars:
@@ -210,14 +241,47 @@ void XrdPosixConfig::initEnv(XrdOucEnv &theEnv, const char *vName, long long &De
       }
 }
   
+/******************************************************************************/
+/*                                O p e n F C                                 */
+/******************************************************************************/
+
+bool XrdPosixConfig::OpenFC(const char *path, int oflag, mode_t mode,
+                            XrdPosixInfo &Info)
+{
+   int rc = XrdPosixXrootd::Open(path, oflag, mode, Info.cbP, &Info);
+
+// Check if we actually can open the file directly via the cache
+//
+   if (rc == -3)
+      {if (*Info.cachePath && errno == 0 && Info.ffReady) return true;
+       rc = -1;
+       if (!errno) errno = ENOPROTOOPT;
+      }
+
+// Return actual result
+//
+   Info.fileFD = rc;
+   return false;
+}
+  
 /******************************************************************************/
 /*                             S e t C o n f i g                              */
 /******************************************************************************/
   
-void XrdPosixConfig::SetConfig(XrdOucPsx &parms)
+bool XrdPosixConfig::SetConfig(XrdOucPsx &parms)
 {
    XrdOucTList *tP;
 
+// Set log routing
+//
+   XrdPosixGlobals::Trace.SetLogger(parms.theLogger);
+   XrdPosixGlobals::theLogger = parms.theLogger;
+
+// Create an error object if we have a logger
+//
+   if (parms.theLogger)
+      XrdPosixGlobals::eDest = new XrdSysError(parms.theLogger, "Posix");
+
 // Set networking mode
 //
    SetIPV4(parms.useV4);
@@ -257,6 +321,8 @@ void XrdPosixConfig::SetConfig(XrdOucPsx &parms)
         if (parms.theCache2)
            {XrdPosixGlobals::myCache2 = parms.theCache2;
             XrdPosixGlobals::theCache = parms.theCache2;
+            if (parms.initCCM) return initCCM(parms);
+            return true;
            }
    else if (parms.theCache)
            {char ebuf[] = {0};
@@ -264,6 +330,8 @@ void XrdPosixConfig::SetConfig(XrdOucPsx &parms)
             initEnv(ebuf);
            }
    else if (parms.mCache && *parms.mCache) initEnv(parms.mCache);
+
+   return true;
 }
 
 /******************************************************************************/
@@ -323,16 +391,6 @@ void XrdPosixConfig::SetIPV4(bool usev4)
 //
    env->PutString((std::string)"NetworkStack", (const std::string)ipmode);
 }
-  
-/******************************************************************************/
-/*                             s e t L o g g e r                              */
-/******************************************************************************/
-
-void XrdPosixConfig::setLogger(XrdSysLogger *logP)
-{
-    XrdPosixGlobals::Trace.SetLogger(logP);
-    XrdPosixGlobals::theLogger = logP;
-}
 
 /******************************************************************************/
 /*                               s e t O i d s                                */
